#!/usr/bin/env python
# coding utf-8

import asyncio
import json
import socket as pysocket
from logging import getLogger
from random import randint
from time import perf_counter
from typing import List, Literal

import numpy as np
from monty.json import jsanitize
from monty.serialization import MontyDecoder

from maggma.cli.multiprocessing import multi
from maggma.cli.settings import CLISettings
from maggma.core import Builder
from maggma.utils import Timeout, tqdm

try:
    import pika
except ImportError:
    raise ImportError("Both pika and aio-pika are required to use RabbitMQ as a broker")

settings = CLISettings()


def find_port():
    sock = pysocket.socket()
    sock.bind(("", 0))
    return sock.getsockname()[1]


def manager(
    url: str,
    builders: List[Builder],
    num_chunks: int,
    num_workers: int,
    queue_prefix: str,
    port: int = 5672,
):
    """
    Rabbit MQ manager for distributed processing that uses a builder prechunk to modify
    the builder and send them out each worker to run.
    """
    logger = getLogger("Manager")

    if not (num_chunks and num_workers):
        raise ValueError("Both num_chunks and num_workers must be non-zero")

    url = url.split("//")[-1]

    logger.info(f"Binding to Manager URL {url}:{port}")

    # Setup connection to RabbitMQ and ensure on all queues is one unit
    connection, channel, status_queue, worker_queue = setup_rabbitmq(url, queue_prefix, port, "work")

    workers = {}  # type: ignore

    logger.debug("Manager started and looking for workers")

    for builder in builders:
        logger.info(f"Working on {builder.__class__.__name__}")
        builder_dict = builder.as_dict()

        try:
            builder.connect()
            chunk_dicts = [{"chunk": d, "distributed": False, "completed": False} for d in builder.prechunk(num_chunks)]
            pbar_distributed = tqdm(
                total=len(chunk_dicts),
                desc=f"Distributed chunks for {builder.__class__.__name__}",
            )

            pbar_completed = tqdm(
                total=len(chunk_dicts),
                desc=f"Completed chunks for {builder.__class__.__name__}",
            )

            logger.info(f"Distributing {len(chunk_dicts)} chunks to workers")

        except NotImplementedError:
            attempt_graceful_shutdown(connection, workers, channel, worker_queue)
            raise RuntimeError(f"Can't distribute process {builder.__class__.__name__} as no prechunk method exists.")

        completed = False

        while not completed:
            completed = all(chunk["completed"] for chunk in chunk_dicts)

            if num_workers <= 0:
                connection.close()
                raise RuntimeError("No workers to distribute chunks to")

            # If workers send messages decode and figure out what do

            _, _, body = channel.basic_get(queue=status_queue, auto_ack=True)

            if body is not None:
                msg = body.decode("utf-8")
                identity = msg.split("_")[-1]

                if "READY" in msg:
                    if identity not in workers:
                        logger.debug(f"Got connection from worker: {msg.split('_')[1]}")
                        workers[identity] = {
                            "working": False,
                            "heartbeats": 1,
                            "last_ping": perf_counter(),
                            "work_index": -1,
                        }

                elif "DONE" in msg:
                    workers[identity]["working"] = False
                    work_ind = workers[identity]["work_index"]
                    if work_ind != -1:
                        chunk_dicts[work_ind]["completed"] = True  # type: ignore
                        pbar_completed.update(1)

                elif "ERROR" in msg:
                    # Remove worker and requeue work sent to it
                    attempt_graceful_shutdown(connection, workers, channel, worker_queue)
                    raise RuntimeError(
                        "At least one worker has stopped with error message: {}".format(msg.split("_")[1])
                    )

                elif "PING" in msg:
                    # Heartbeat from worker (no pong response)
                    workers[identity]["last_ping"] = perf_counter()
                    workers[identity]["heartbeats"] += 1

            # Decide if any workers are dead and need to be removed
            handle_dead_workers(connection, workers, channel, worker_queue)

            for work_index, chunk_dict in enumerate(chunk_dicts):
                if not chunk_dict["distributed"]:
                    temp_builder_dict = dict(**builder_dict)
                    temp_builder_dict.update(chunk_dict["chunk"])  # type: ignore
                    temp_builder_dict = jsanitize(temp_builder_dict)

                    # Send work for available workers
                    for identity in workers:
                        if not workers[identity]["working"]:
                            # Send out a chunk to idle worker
                            channel.basic_publish(
                                exchange="",
                                routing_key=worker_queue,
                                body=json.dumps(temp_builder_dict).encode("utf-8"),
                            )

                            workers[identity]["work_index"] = work_index
                            workers[identity]["working"] = True
                            chunk_dicts[work_index]["distributed"] = True
                            pbar_distributed.update(1)

    # Send EXIT to any remaining workers
    logger.info("Sending exit messages to workers once they are done")
    attempt_graceful_shutdown(connection, workers, channel, worker_queue)


def setup_rabbitmq(url: str, queue_prefix: str, port: int, outbound_queue: Literal["status", "work"]):
    connection = pika.BlockingConnection(pika.ConnectionParameters(url, port))
    channel = connection.channel()
    channel.basic_qos(prefetch_count=1, global_qos=True)

    # Ensure both worker status and work distribution queues exist
    status_queue = queue_prefix + "_status"
    worker_queue = queue_prefix + "_work"

    channel.queue_declare(queue=status_queue, auto_delete=True)
    channel.queue_declare(queue=worker_queue, auto_delete=True)

    # Clear out outbound queue
    if outbound_queue == "work":
        channel.queue_purge(queue=worker_queue)
    else:
        channel.queue_purge(queue=status_queue)

    return connection, channel, status_queue, worker_queue


def attempt_graceful_shutdown(connection, workers, channel, worker_queue):
    for _ in workers:
        channel.basic_publish(
            exchange="",
            routing_key=worker_queue,
            body=b"EXIT",
        )
    connection.close()


def handle_dead_workers(connection, workers, channel, worker_queue):
    if len(workers) == 1:
        # Use global timeout
        identity = next(iter(workers.keys()))
        if (perf_counter() - workers[identity]["last_ping"]) >= settings.WORKER_TIMEOUT:
            attempt_graceful_shutdown(connection, workers, channel, worker_queue)
            raise RuntimeError("Worker has timed out. Stopping distributed build.")

    elif len(workers) == 2:
        # Use 10% ratio between workers
        workers_sorted = sorted(workers.items(), key=lambda x: x[1]["heartbeats"])

        ratio = workers_sorted[1][1]["heartbeats"] / workers_sorted[0][1]["heartbeats"]

        if ratio <= 0.1:
            attempt_graceful_shutdown(connection, workers, channel, worker_queue)
            raise RuntimeError("One worker has timed out. Stopping distributed build.")

    elif len(workers) > 2:
        # Calculate modified z-score of heartbeat counts and see if any are <= -3.5
        hearbeat_vals = [w["heartbeats"] for w in workers.values()]
        median = np.median(hearbeat_vals)
        mad = np.median([abs(i - median) for i in hearbeat_vals])
        if mad > 0:
            for identity in list(workers.keys()):
                z_score = 0.6745 * (workers[identity]["heartbeats"] - median) / mad
                if z_score <= -3.5:
                    attempt_graceful_shutdown(connection, workers, channel, worker_queue)
                    raise RuntimeError("At least one worker has timed out. Stopping distributed build.")


def worker(url: str, port: int, num_processes: int, no_bars: bool, queue_prefix: str):
    """
    Simple distributed worker that connects to a manager asks for work and deploys
    using multiprocessing
    """
    identity = f"{randint(0, 0x10000):04X}-{randint(0, 0x10000):04X}"
    logger = getLogger(f"Worker {identity}")

    url = url.split("//")[-1]

    logger.info(f"Connecting to Manager at {url}:{port}")

    # Setup connection to RabbitMQ and ensure on all queues is one unit
    connection, channel, status_queue, worker_queue = setup_rabbitmq(url, queue_prefix, port, "status")

    # Send ready signal to status queue
    channel.basic_publish(
        exchange="",
        routing_key=status_queue,
        body=f"READY_{identity}".encode(),
    )

    try:
        running = True
        while running:
            # Wait for work from manager
            with Timeout(seconds=settings.MANAGER_TIMEOUT):
                _, _, body = channel.basic_get(queue=worker_queue, auto_ack=True)

            if body is not None:
                message = body.decode("utf-8")

                if "@class" in message and "@module" in message:
                    # We have a valid builder
                    work = json.loads(message)
                    builder = MontyDecoder().process_decoded(work)

                    logger.info(f"Working on builder {builder.__class__}")

                    channel.basic_publish(
                        exchange="",
                        routing_key=status_queue,
                        body=f"WORKING_{identity}".encode(),
                    )
                    work = json.loads(message)
                    builder = MontyDecoder().process_decoded(work)

                    asyncio.run(
                        multi(
                            builder,
                            num_processes,
                            no_bars=no_bars,
                            heartbeat_func=ping_manager,
                            heartbeat_func_kwargs={
                                "channel": channel,
                                "identity": identity,
                                "status_queue": status_queue,
                            },
                        )
                    )

                    channel.basic_publish(
                        exchange="",
                        routing_key=status_queue,
                        body=f"DONE_{identity}".encode(),
                    )

                elif message == "EXIT":
                    # End the worker
                    running = False

    except Exception as e:
        logger.error(f"A worker failed with error: {e!r}")
        channel.basic_publish(
            exchange="",
            routing_key=status_queue,
            body=f"ERROR_{identity}".encode(),
        )
        connection.close()

    connection.close()


def ping_manager(channel, identity, status_queue):
    channel.basic_publish(
        exchange="",
        routing_key=status_queue,
        body=f"PING_{identity}".encode(),
    )



from typing import Dict

from fastapi import HTTPException, Query

from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS


class PaginationQuery(QueryOperator):
    """Query operators to provides Pagination"""

    def __init__(self, default_limit: int = 100, max_limit: int = 1000):
        """
        Args:
            default_limit: the default number of documents to return
            max_limit: max number of documents to return
        """

        self.default_limit = default_limit
        self.max_limit = max_limit

        def query(
            _page: int = Query(
                None,
                description="Page number to request (takes precedent over _limit and _skip).",
            ),
            _per_page: int = Query(
                default_limit,
                description="Number of entries to show per page (takes precedent over _limit and _skip)."
                f" Limited to {max_limit}.",
            ),
            _skip: int = Query(
                0,
                description="Number of entries to skip in the search.",
            ),
            _limit: int = Query(
                default_limit,
                description=f"Max number of entries to return in a single query. Limited to {max_limit}.",
            ),
        ) -> STORE_PARAMS:
            """
            Pagination parameters for the API Endpoint
            """

            if _page is not None:
                if _per_page > max_limit:
                    raise HTTPException(
                        status_code=400,
                        detail="Requested more data per query than allowed by this endpoint."
                        f" The max limit is {max_limit} entries",
                    )

                if _page < 0 or _per_page < 0:
                    raise HTTPException(
                        status_code=400,
                        detail="Cannot request negative _page or _per_page values",
                    )

                return {
                    "skip": ((_page - 1) * _per_page) if _page >= 1 else 0,
                    "limit": _per_page,
                }

            else:
                if _limit > max_limit:
                    raise HTTPException(
                        status_code=400,
                        detail="Requested more data per query than allowed by this endpoint."
                        f" The max limit is {max_limit} entries",
                    )

                if _skip < 0 or _limit < 0:
                    raise HTTPException(
                        status_code=400,
                        detail="Cannot request negative _skip or _limit values",
                    )

                return {"skip": _skip, "limit": _limit}

        self.query = query  # type: ignore

    def query(self):
        "Stub query function for abstract class"

    def meta(self) -> Dict:
        """
        Metadata for the pagination params
        """
        return {"max_limit": self.max_limit}



#!/usr/bin/env python


from pathlib import Path

from setuptools import find_packages, setup

module_dir = Path(__file__).resolve().parent

with open(module_dir / "README.md") as f:
    long_desc = f.read()
setup(
    name="maggma",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Framework to develop datapipelines from files on disk to full dissemination API",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/materialsproject/maggma",
    author="The Materials Project",
    author_email="feedback@materialsproject.org",
    license="modified BSD",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"maggma": ["py.typed"]},
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
        "setuptools",
        "ruamel.yaml<0.18",
        "pydantic>=2.0",
        "pydantic-settings>=2.0.3",
        "pymongo>=4.2.0",
        "monty>=2023.9.25",
        "mongomock>=3.10.0",
        "pydash>=4.1.0",
        "jsonschema>=3.1.1",
        "tqdm>=4.19.6",
        "mongogrant>=0.3.1",
        "aioitertools>=0.5.1",
        "numpy>=1.17.3",
        "fastapi>=0.42.0",
        "pyzmq>=24.0.1",
        "dnspython>=1.16.0",
        "sshtunnel>=0.1.5",
        "msgpack>=0.5.6",
        "orjson>=3.9.0",
        "boto3>=1.20.41",
        "python-dateutil>=2.8.2",
        "uvicorn>=0.18.3",
    ],
    extras_require={
        "vault": ["hvac>=0.9.5"],
        "memray": ["memray>=1.7.0"],
        "montydb": ["montydb>=2.3.12"],
        "notebook_runner": ["IPython>=8.11", "nbformat>=5.0", "regex>=2020.6"],
        "azure": ["azure-storage-blob>=12.16.0", "azure-identity>=1.12.0"],
        "testing": [
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "pytest-asyncio",
            "pytest-xdist",
            "pre-commit",
            "moto",
            "ruff",
            "responses<0.22.0",
            "types-pyYAML",
            "types-setuptools",
            "types-python-dateutil",
            "starlette[full]",
        ],
        "docs": [
            "mkdocs>=1.4.0",
            "mkdocs-material>=8.3.9",
            "mkdocs-minify-plugin>=0.5.0",
            "mkdocstrings[python]>=0.18.1",
            "jinja2<3.2.0",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "Operating System :: OS Independent",
        "Topic :: Other/Nonlisted Topic",
        "Topic :: Database :: Front-Ends",
        "Topic :: Scientific/Engineering",
    ],
    entry_points={"console_scripts": ["mrun = maggma.cli:run"]},
)



from datetime import datetime
from random import randint

import pytest
from fastapi import FastAPI
from maggma.api.resource import PostOnlyResource
from maggma.stores import MemoryStore
from pydantic import BaseModel, Field
from starlette.testclient import TestClient


class Owner(BaseModel):
    name: str = Field(..., title="Owner's name")
    age: int = Field(None, title="Owne'r Age")
    weight: float = Field(None, title="Owner's weight")
    last_updated: datetime = Field(None, title="Last updated date for this record")


owners = (
    [Owner(name=f"Person{i}", age=i + 3, weight=100 + i) for i in list(range(10))]
    + [Owner(name="PersonAge9", age=9, weight=float(randint(155, 195)))]
    + [Owner(name="PersonWeight150", age=randint(10, 15), weight=float(150))]
    + [Owner(name="PersonAge20Weight200", age=20, weight=float(200))]
)

total_owners = len(owners)


@pytest.fixture()
def owner_store():
    store = MemoryStore("owners", key="name")
    store.connect()
    store.update([d.dict() for d in owners])
    return store


def test_init(owner_store):
    resource = PostOnlyResource(store=owner_store, model=Owner)
    assert len(resource.router.routes) == 2


def test_msonable(owner_store):
    owner_resource = PostOnlyResource(store=owner_store, model=Owner)
    endpoint_dict = owner_resource.as_dict()

    for k in ["@class", "@module", "store", "model"]:
        assert k in endpoint_dict

    assert isinstance(endpoint_dict["model"], str)
    assert endpoint_dict["model"] == "tests.api.test_post_resource.Owner"


def test_post_to_search(owner_store):
    endpoint = PostOnlyResource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    assert client.post("/").status_code == 200


@pytest.mark.xfail()
def test_problem_query_params(owner_store):
    endpoint = PostOnlyResource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    client.get("/?param=test").status_code



from typing import Callable, Dict, List

from fastapi import Depends, Request, Response

from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS, attach_signature
from maggma.core.store import Store


def attach_query_ops(
    function: Callable[[List[STORE_PARAMS]], Dict], query_ops: List[QueryOperator]
) -> Callable[[List[STORE_PARAMS]], Dict]:
    """
    Attach query operators to API compliant function
    The function has to take a list of STORE_PARAMs as the only argument

    Args:
        function: the function to decorate
    """
    attach_signature(
        function,
        annotations={
            **{f"dep{i}": STORE_PARAMS for i, _ in enumerate(query_ops)},
            "request": Request,
            "temp_response": Response,
        },
        defaults={f"dep{i}": Depends(dep.query) for i, dep in enumerate(query_ops)},
    )
    return function


def generate_query_pipeline(query: dict, store: Store):
    """
    Generate the generic aggregation pipeline used in GET endpoint queries

    Args:
        query: Query parameters
        store: Store containing endpoint data
    """
    pipeline = [
        {"$match": query["criteria"]},
    ]

    sorting = query.get("sort", False)

    if sorting:
        sort_dict = {"$sort": {}}  # type: dict
        sort_dict["$sort"].update(query["sort"])
        sort_dict["$sort"].update({store.key: 1})  # Ensures sort by key is last in dict to fix determinacy

    projection_dict = {"_id": 0}  # Do not return _id by default

    if query.get("properties", False):
        projection_dict.update({p: 1 for p in query["properties"]})

    if sorting:
        pipeline.append(sort_dict)

    pipeline.append({"$project": projection_dict})
    pipeline.append({"$skip": query["skip"] if "skip" in query else 0})

    if query.get("limit", False):
        pipeline.append({"$limit": query["limit"]})

    return pipeline



from datetime import datetime
from typing import Generic, List, Optional, TypeVar

from pydantic import BaseModel, Field, validator

from maggma import __version__

""" Describes the Materials API Response """


DataT = TypeVar("DataT")


class Meta(BaseModel):

    """
    Meta information for the MAPI Response
    """

    api_version: str = Field(
        __version__,
        description="a string containing the version of the Materials API implementation, e.g. v0.9.5",
    )

    time_stamp: datetime = Field(
        description="a string containing the date and time at which the query was executed",
        default_factory=datetime.utcnow,
    )

    total_doc: Optional[int] = Field(None, description="the total number of documents available for this query", ge=0)

    class Config:
        extra = "allow"


class Error(BaseModel):
    """
    Base Error model for General API
    """

    code: int = Field(..., description="The error code")
    message: str = Field(..., description="The description of the error")

    @classmethod
    def from_traceback(cls, traceback):
        pass


class Response(BaseModel, Generic[DataT]):
    """
    A Generic API Response
    """

    data: Optional[List[DataT]] = Field(None, description="List of returned data")
    errors: Optional[List[Error]] = Field(None, description="Any errors on processing this query")
    meta: Optional[Meta] = Field(None, description="Extra information for the query")

    @validator("errors", always=True)
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v

    @validator("meta", pre=True, always=True)
    def default_meta(cls, v, values):
        if v is None:
            v = Meta().dict()
        if v.get("total_doc", None) is None:
            if values.get("data", None) is not None:
                v["total_doc"] = len(values["data"])
            else:
                v["total_doc"] = 0
        return v


class S3URLDoc(BaseModel):

    """
    S3 pre-signed URL data returned by the S3 URL resource
    """

    url: str = Field(
        ...,
        description="Pre-signed download URL",
    )

    requested_datetime: datetime = Field(..., description="Datetime for when URL was requested")

    expiry_datetime: datetime = Field(..., description="Expiry datetime of the URL")



#!/usr/bin/env python
# coding utf-8


import asyncio
import logging
import sys
from datetime import datetime
from itertools import chain

import click
from monty.serialization import loadfn

from maggma.cli.distributed import find_port
from maggma.cli.multiprocessing import multi
from maggma.cli.serial import serial
from maggma.cli.settings import CLISettings
from maggma.cli.source_loader import ScriptFinder, load_builder_from_source
from maggma.utils import ReportingHandler, TqdmLoggingHandler

sys.meta_path.append(ScriptFinder())

settings = CLISettings()


@click.command()
@click.argument("builders", nargs=-1, type=click.Path(exists=True), required=True)
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Controls logging level per number of v's",
    default=0,
)
@click.option(
    "-n",
    "--num-processes",
    "num_processes",
    help="Number of processes to spawn for each worker. Defaults to single processing",
    default=1,
    type=click.IntRange(1),
)
@click.option(
    "-r",
    "--reporting",
    "reporting_store",
    help="Store in JSON/YAML form to send reporting data to",
    type=click.Path(exists=True),
)
@click.option("-u", "--url", "url", default=None, type=str, help="URL for the distributed manager")
@click.option(
    "-p",
    "--port",
    "port",
    default=None,
    type=int,
    help="Port for distributed communication. mrun will find an open port if None is provided to the manager",
)
@click.option(
    "-N",
    "--num-chunks",
    "num_chunks",
    default=0,
    type=int,
    help="Number of chunks to distribute to workers",
)
@click.option(
    "-w",
    "--num-workers",
    "num_workers",
    default=0,
    type=int,
    help="Number of distributed workers to process chunks",
)
@click.option("--no_bars", is_flag=True, help="Turns of Progress Bars for headless operations")
@click.option("--rabbitmq", is_flag=True, help="Enables the use of RabbitMQ as the work broker")
@click.option(
    "-q",
    "--queue_prefix",
    "queue_prefix",
    default="builder",
    type=str,
    help="Prefix to use in queue names when RabbitMQ is select as the broker",
)
@click.option(
    "-m",
    "--memray",
    "memray",
    default=False,
    type=bool,
    help="Option to profile builder memory usage with Memray",
)
@click.option(
    "-md",
    "--memray-dir",
    "memray_dir",
    default=None,
    type=str,
    help="""Directory to dump memory profiler output files. Only runs if --memray is True.
    Will create directory if directory does not exist, mimicking mkdir -p command.
    If not provided files will be dumped to system's temp directory""",
)
@click.pass_context
def run(
    ctx,
    builders,
    verbosity,
    reporting_store,
    num_workers,
    url,
    port,
    num_chunks,
    no_bars,
    num_processes,
    rabbitmq,
    queue_prefix,
    memray,
    memray_dir,
    memray_file=None,
    follow_fork=False,
):
    # Import profiler and setup directories to dump profiler output
    if memray:
        from memray import FileDestination, Tracker

        if memray_dir:
            import os

            os.makedirs(memray_dir, exist_ok=True)

            memray_file = f"{memray_dir}/{builders[0]}_{datetime.now().isoformat()}.bin"
        else:
            memray_file = f"{settings.TEMP_DIR}/{builders[0]}_{datetime.now().isoformat()}.bin"

        if num_processes > 1:
            follow_fork = True

        # Click context manager handles creation and clean up of profiler dump files for memray tracker
        ctx.obj = ctx.with_resource(
            Tracker(
                destination=FileDestination(memray_file),
                native_traces=False,
                trace_python_allocators=False,
                follow_fork=follow_fork,
            )
        )

    # Import proper manager and worker
    if rabbitmq:
        from maggma.cli.rabbitmq import manager, worker
    else:
        from maggma.cli.distributed import manager, worker

    # Set Logging
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(len(levels) - 1, verbosity)]  # capped to number of levels
    root = logging.getLogger()
    root.setLevel(level)
    ch = TqdmLoggingHandler()
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    root.addHandler(ch)

    builder_objects = []

    for b in builders:
        if str(b).endswith(".py") or str(b).endswith(".ipynb"):
            builder_objects.append(load_builder_from_source(b))
        else:
            builder_objects.append(loadfn(b))

    builder_objects = [b if isinstance(b, list) else [b] for b in builder_objects]
    builder_objects = list(chain.from_iterable(builder_objects))

    if reporting_store:
        reporting_store = loadfn(reporting_store)
        root.addHandler(ReportingHandler(reporting_store))

    if url:
        if num_chunks > 0:
            # Manager
            if port is None:
                port = find_port()
                root.critical(f"Using random port for mrun manager: {port}")

            if rabbitmq:
                manager(
                    url=url,
                    port=port,
                    builders=builder_objects,
                    num_chunks=num_chunks,
                    num_workers=num_workers,
                    queue_prefix=queue_prefix,
                )
            else:
                manager(
                    url=url,
                    port=port,
                    builders=builder_objects,
                    num_chunks=num_chunks,
                    num_workers=num_workers,
                )

        else:
            # Worker
            if rabbitmq:
                worker(
                    url=url,
                    port=port,
                    num_processes=num_processes,
                    no_bars=no_bars,
                    queue_prefix=queue_prefix,
                )
            else:
                worker(url=url, port=port, num_processes=num_processes, no_bars=no_bars)
    else:
        if num_processes == 1:
            for builder in builder_objects:
                serial(builder, no_bars)
        else:
            loop = asyncio.get_event_loop()
            for builder in builder_objects:
                loop.run_until_complete(multi(builder=builder, num_processes=num_processes, no_bars=no_bars))

    if memray_file:
        import subprocess

        subprocess.run(["memray", "flamegraph", memray_file], shell=False)
